from abc import ABCMeta, abstractmethod

import numpy
from catsim import irt
from scipy.optimize import fmin
from scipy.optimize import differential_evolution


class Estimator:
    """Base class for proficiency estimators"""

    __metaclass__ = ABCMeta

    def __init__(self):
        super(Estimator, self).__init__()

    @abstractmethod
    def estimate(
        self, response_vector: list, administered_items: numpy.ndarray, current_theta: float
    ) -> float:
        """Uses a hill-climbing algorithm to find and return the theta value
        that maximizes the likelihood function, given a response vector and a
        matrix with the administered items parameters.

        :param response_vector: a binary list containing the response vector
        :param administered_items: a matrix containing the parameters of the
                                   answered items
        :param current_theta: the current estimation of the examinee's :math:`\\theta` value
        :returns: a new estimation of the examinee's proficiency, given his answers up until now
        """
        pass


class HillClimbingEstimator(Estimator):
    """Estimator that uses a hill-climbing algorithm to maximize the likelihood function

    :param precision: number of decimal points of precision
    :param verbose: verbosity level of the maximization method
    """

    def __init__(self, precision: int=6, verbose: bool=False):
        super(HillClimbingEstimator, self).__init__()
        self._precision = precision
        self._verbose = verbose
        self._evaluations = 0
        self._calls = 0

    @property
    def calls(self):
        """How many times the estimator has been called to maximize/minimize the log-likelihood function

        :returns: number of times the estimator has been called to maximize/minimize the log-likelihood function"""
        return self._calls

    @property
    def evaluations(self):
        """Total number of times the estimator has evaluated the log-likelihood function during its existence

        :returns: number of function evaluations"""
        return self._evaluations

    @property
    def avg_evaluations(self):
        """Average number of function evaluations for all tests the estimator has been used

        :returns: average number of function evaluations"""
        return self._evaluations / self._calls

    def estimate(
        self,
        response_vector: list,
        administered_items: numpy.ndarray,
        current_theta: float=None
    ) -> float:
        """Uses a hill-climbing algorithm to find and returns the theta value
        that maximizes the likelihood function, given a response vector and a
        matrix with the administered items parameters.

        :param response_vector: a binary list containing the response vector
        :param administered_items: a matrix containing the parameters of the
                                   answered items
        :param current_theta: not used by this selector
        :returns: a new estimation of the examinee's proficiency, given his answers up until now
        :rtype: float
        """

        self._calls += 1

        if set(response_vector) == 1:
            return float('inf')
        elif set(response_vector) == 0:
            return float('-inf')

        lbound = min(administered_items[:, 1])
        ubound = max(administered_items[:, 1])
        best_theta = -float('inf')
        max_ll = -float('inf')

        self._evaluations = 0

        for _ in range(10):
            intervals = numpy.linspace(lbound, ubound, 10)
            if self._verbose:
                print('Bounds: ' + str(lbound) + ' ' + str(ubound))
                print('Interval size: ' + str(intervals[1] - intervals[0]))

            for ii in intervals:
                self._evaluations += 1
                ll = irt.logLik(ii, response_vector, administered_items)
                if ll > max_ll:
                    max_ll = ll

                    if self._verbose:
                        print(
                            'Iteration: {0}, Theta: {1}, LL: {2}'.format(
                                self._evaluations, ii, ll
                            )
                        )

                    if abs(best_theta - ii) < float('1e-' + str(self._precision)):
                        return ii

                    best_theta = ii

                else:
                    lbound = best_theta - (intervals[1] - intervals[0])
                    ubound = ii
                    break

        return best_theta


class FMinEstimator(Estimator):
    """Estimator that uses :py:func:`scipy.optimize.fmin` to minimize the negative log-likelihood function"""

    def __init__(self):
        super(FMinEstimator, self).__init__()
        self._evaluations = 0
        self._calls = 0

    @property
    def calls(self):
        """How many times the estimator has been called to maximize/minimize the log-likelihood function

        :returns: number of times the estimator has been called to maximize/minimize the log-likelihood function"""
        return self._calls

    @property
    def evaluations(self):
        """Total number of times the estimator has evaluated the log-likelihood function during its existence

        :returns: number of function evaluations"""
        return self._evaluations

    @property
    def avg_evaluations(self):
        """Average number of function evaluations for all tests the estimator has been used

        :returns: average number of function evaluations"""
        return self._evaluations / self._calls

    def estimate(
        self, response_vector: list, administered_items: numpy.ndarray, current_theta: float
    ) -> float:
        """Uses :py:func:`scipy.optimize.fmin` to find and return the theta value
        that minimizes the negative log-likelihood function, given a response vector, a
        matrix with the administered items parameters and the current :math:`\\theta`.

        :param response_vector: a binary list containing the response vector
        :param administered_items: a matrix containing the parameters of the
                                   answered items
        :param current_theta: the current estimation of the examinee's :math:`\\theta` value
        :returns: a new estimation of the examinee's proficiency, given his answers up until now
        """
        self._calls += 1
        res = fmin(
            irt.negativelogLik,
            current_theta,
            (response_vector, administered_items),
            disp=False,
            full_output=True
        )

        self._evaluations = res[3]
        return res[0][0]


class DifferentialEvolutionEstimator(Estimator):
    """Estimator that uses :py:func:`scipy.optimize.differential_evolution` to minimize the negative log-likelihood function

    :param bounds: a tuple containing both lower and upper bounds for the differential
                   evolution algorithm search space. In theory, it is best if they
                   represent the minimum and maximum possible :math:`\\theta` values;
                   in practice, one could also use the smallest and largest difficulty
                   parameters in the item bank, in case no better bounds for
                   :math:`\\theta` exist.
    """

    def __init__(self, bounds: tuple):
        super(DifferentialEvolutionEstimator, self).__init__()
        self._lower_bound = min(bounds)
        self._upper_bound = max(bounds)
        self._evaluations = 0
        self._calls = 0

    @property
    def calls(self):
        """How many times the estimator has been called to maximize/minimize the log-likelihood function

        :returns: number of times the estimator has been called to maximize/minimize the log-likelihood function"""
        return self._calls

    @property
    def evaluations(self):
        """Total number of times the estimator has evaluated the log-likelihood function during its existence

        :returns: number of function evaluations"""
        return self._evaluations

    @property
    def avg_evaluations(self):
        """Average number of function evaluations for all tests the estimator has been used

        :returns: average number of function evaluations"""
        return self._evaluations / self._calls

    def estimate(
        self,
        response_vector: list,
        administered_items: numpy.ndarray,
        current_theta: float=None
    ) -> float:
        """Uses :py:func:`scipy.optimize.fmin` to find and return the theta value
        that minimizes the negative log-likelihood function, given a response vector, a
        matrix with the administered items parameters and the current :math:`\\theta`.

        :param response_vector: a binary list containing the response vector
        :param administered_items: a matrix containing the parameters of the
                                   answered items
        :param current_theta: the current estimation of the examinee's :math:`\\theta` value
        :returns: not used by this selector
        """

        self._calls += 1

        res = differential_evolution(
            irt.negativelogLik,
            bounds=[
                [self._lower_bound * 2, self._upper_bound * 2]
            ],
            args=(response_vector, administered_items)
        )

        self._evaluations = res.nfev

        return res.x[0]



"""Module containing functions relevant to the process of simulating the
application of adaptive tests. Most of this module is based on the work of
[Bar10]_."""

import time
import numpy
from catsim import irt, cat
from catsim.initialization import Initializer
from catsim.selection import Selector
from catsim.estimation import Estimator
from catsim.stopping import Stopper


class Simulator:
    """Class representing the simulator. It gathers several objects that describe the full
    simulation process and simulates one or more computerized adaptive tests

    :param items: a matrix containing item parameters
    :param examinees: an integer with the number of examinees, whose real :math:`\\theta` values will be
                      sampled from a normal distribution; or a :py:type:list containing said
                      :math:`\\theta_0` values
    """

    def __init__(self, items: numpy.ndarray, examinees):
        irt.validate_item_bank(items)

        # adds a column for each item's exposure rate and their cluster membership
        items = numpy.append(items, numpy.zeros([items.shape[0], 1]), axis=1)

        self._duration = 0
        self._items = items
        self._estimations = []
        self._administered_items = []

        # `examinees` is passed to its special setter
        self.examinees = examinees

    @property
    def items(self) -> numpy.ndarray:
        """Item matrix used by the simulator. If the simulation already
        occurred, a column containin item esposure rates will be added to the
        matrix."""
        return self._items

    @property
    def administered_items(self) -> list:
        """List of lists containin the indexes of items administered to each
        examinee during the simulation."""
        return self._administered_items

    @property
    def all_estimations(self) -> list:
        """List of lists containing all estimated :math:`\\hat\\theta` values
        for all examinees during each step of the test."""
        return self._estimations

    @property
    def estimations(self) -> list:
        """Final estimated :math:`\\hat\\theta` values for all examinees."""
        return [ests[-1] for ests in self._estimations]

    @property
    def examinees(self) -> list:
        """List containing examinees true proficiency values (:math:`\\theta`)."""
        return self._examinees

    @property
    def duration(self) -> float:
        """Duration of the simulation, in milliseconds."""
        return self._duration

    @property
    def bias(self) -> float:
        """Bias between the estimated and true proficiencies. This property is only available after :py:func:`simulate` has been successfully called. For more information on estimation bias, see :py:func:`catsim.cat.bias`"""
        return self._bias

    @property
    def mse(self) -> float:
        """Mean-squared error between the estimated and true proficiencies. This
        property is only available after :py:func:`simulate` has been successfully
        called. For more information on the mean-squared error of estimation, see
        :py:func:`catsim.cat.mse`"""
        return self._mse

    @property
    def rmse(self) -> float:
        """Root mean-squared error between the estimated and true proficiencies. This
        property is only available after :py:func:`simulate` has been successfully
        called. For more information on the root mean-squared error of estimation, see
        :py:func:`catsim.cat.rmse`"""
        return self._rmse

    @examinees.setter
    def examinees(self, examinees):
        if type(examinees) == int:
            self._examinees = numpy.random.normal(0, 1, examinees)
        elif type(examinees) == list:
            self._examinees = numpy.array(examinees)

    def simulate(
        self,
        initializer: Initializer,
        selector: Selector,
        estimator: Estimator,
        stopper: Stopper,
        verbose: bool=False
    ):
        """Simulates a computerized adaptive testing application to one or more examinees

        :param initializer: an initializer that selects examinees :math:`\\theta_0`
        :param selector: a selector that selects new items to be presented to examinees
        :param estimator: an estimator that reestimates examinees proficiencies after each item is applied
        :param stopper: an object with a stopping criteria for the test
        :param verbose: whether to periodically print a message regarding the progress of the simulation. Good for longer simulations.

        >>> from catsim.initialization import RandomInitializer
        >>> from catsim.selection import MaxInfoSelector
        >>> from catsim.estimation import HillClimbingEstimator
        >>> from catsim.stopping import MaxItemStopper
        >>> from catsim.cat import generate_item_bank
        >>> initializer = RandomInitializer()
        >>> selector = MaxInfoSelector()
        >>> estimator = HillClimbingEstimator()
        >>> stopper = MaxItemStopper(20)
        >>> Simulator(generate_item_bank(100), 10).simulate(initializer, selector, estimator, stopper)
        """

        # if verbose:
        print(
            'Starting simulation: {0} {1} {2} {3}'.format(
                initializer.__class__, selector.__class__, estimator.__class__, stopper.__class__
            )
        )
        start_time = int(round(time.time() * 1000))
        for current_examinee, true_theta in enumerate(self.examinees):

            # if verbose:
            print('{0}/{1} examinees...'.format(current_examinee + 1, len(self.examinees)))

            est_theta = initializer.initialize()
            response_vector, administered_items, est_thetas = [], [], []

            while not stopper.stop(self.items[administered_items], est_thetas):
                try:
                    selected_item = selector.select(self.items, administered_items, est_theta)
                except:
                    print(len(administered_items))
                    raise

                # simulates the examinee's response via the three-parameter
                # logistic function
                response = irt.tpm(
                    true_theta, self.items[selected_item][0], self.items[selected_item][1],
                    self.items[selected_item][2]
                ) >= numpy.random.uniform()

                response_vector.append(response)

                # adds the item selected by the selector to the pool of administered items
                administered_items.append(selected_item)

                # if the response vector is composed only of correct or wrong answers, use dodd's method of estimation
                if len(set(response_vector)) == 1:
                    est_theta = cat.dodd(est_theta, self.items, response)
                else:  # else, use the given estimator
                    est_theta = estimator.estimate(
                        response_vector, self.items[administered_items], est_theta
                    )

                # flatten the list of lists so that we can count occurrences of items easier
                flattened_administered_items = [
                    administered_item
                    for administered_list in self.administered_items
                    for administered_item in administered_list
                ]

                # update the exposure value for this item
                # r = number of tests item has been used / total number of tests
                self.items[selected_item, 3] = numpy.sum(
                    flattened_administered_items == selected_item
                ) / len(self.examinees)

                est_thetas.append(est_theta)

            self._estimations.append(est_thetas)
            self._administered_items.append(administered_items)

        self._duration = int(round(time.time() * 1000)) - start_time

        # if verbose:
        print('Simulation took {0} milliseconds'.format(self._duration))

        self._bias = cat.bias(self.examinees, self.estimations)
        self._mse = cat.mse(self.examinees, self.estimations)
        self._rmse = cat.rmse(self.examinees, self.estimations)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import math
import numpy


def tpm(theta: float, a: float, b: float, c: float=0) -> float:
    """Item Response Theory three-parameter logistic function [Ayala2009]_:

    .. math:: P(X_i = 1| \\theta) = c_i + \\frac{1-c_i}{1+ e^{a_i(\\theta-b_i)}}

    :param theta: the individual's proficiency value. This parameter value has
                  no boundary, but if a distribution of the form :math:`N(0, 1)` was
                  used to estimate the parameters, then :math:`-4 \\leq \\theta \\leq
                  4`.

    :param a: the discrimination parameter of the item, usually a positive
              value in which :math:`0.8 \\leq a \\leq 2.5`.

    :param b: the item difficulty parameter. This parameter value has no
              boundaries, but it is necessary that it be in the same value space
              as `theta` (usually :math:`-4 \\leq b \\leq 4`).

    :param c: the item pseudo-guessing parameter. Being a probability,
              :math:`0\\leq c \\leq 1`, but items considered good usually have
              :math:`c \\leq 0.2`.
    """
    return c + ((1 - c) / (1 + _tpm_exponent(theta, a, b)))


def _w(theta, a, b, c):
    p_star = math.pow(1 + _tpm_exponent(theta, a, b), -1)
    q_star = 1 - p_star
    p = tpm(theta, a, b, c)
    q = 1 - p

    return p_star * q_star / p * q


def _tpm_exponent(theta, a, b):
    return math.exp(-a * (theta - b))


def _theta_map(
    theta: float,
    response_vector: list,
    administered_items: numpy.ndarray,
    mu: float=0,
    sigma: float=1
) -> float:
    """ Maximum a posteriori estimation function for :math:`\\hat\\theta` [Andrade2000]_, given by:

    .. math:: \\sum_{i-1}^{I} a_i(1-c_i)(u_{ji}-P_{JI})W_{ji} - \\frac{(\\theta_j - \\mu)}{\\sigma^2}

    where .. math:: W = \\frac{P^*Q^*}{P^Q^}

    where :math:`P` is given by :py:func:`catsim.irt.tpm`, :math:`P^* = (1 + e^{a_i(\\theta-b_i)})^{-1}`,
    :math:`Q^* = 1 - P^*` and :math:`Q = 1 - P`
    """
    return sum(
        [
            administered_items[i, 0] * (1 - administered_items[i, 2]) * (
                response_vector[i] - tpm(
                    theta, administered_items[i, 0], administered_items[
                        i, 1
                    ], administered_items[i, 2]
                ) * _w(
                    theta, administered_items[
                        i, 0
                    ], administered_items[i, 1], administered_items[i, 2]
                )
            ) for i in range(len(response_vector))
        ]
    ) - (theta - mu) / (sigma**2)


def var(theta: float, items: numpy.ndarray) -> float:
    """Computes the variance (:math:`Var`) of the proficiency estimate of a test at a
    specific :math:`\\theta` value [Ayala2009]_:

    .. math:: Var = \\frac{1}{I(\\theta)}

    where :math:`I(\\theta)` is the test information (see :py:func:`test_info`).

    :param theta: a proficiency value.
    :param items: a matrix containing item parameters.
    :returns: the variance of proficency estimation at `theta` for a test represented by `items`.
    """
    try:
        return 1 / test_info(theta, items)
    except ZeroDivisionError:
        return float('-inf')


def see(theta: float, items: numpy.ndarray) -> float:
    """Computes the standard error of estimation (:math:`SEE`) of a test at a
    specific :math:`\\theta` value [Ayala2009]_:

    .. math:: SEE = \\\\sqrt{frac{1}{I(\\theta)}}

    where :math:`I(\\theta)` is the test information (see :py:func:`test_info`).

    :param theta: a proficiency value.
    :param items: a matrix containing item parameters.
    :returns: the standard error of estimation at `theta` for a test represented by `items`.
    """
    try:
        return math.sqrt(var(theta, items))
    except ZeroDivisionError:
        return float('-inf')


def test_info(theta: float, items: numpy.ndarray):
    """Computes the test information of a test at a specific :math:`\\theta` value [Ayala2009]_:

    .. math:: I(\\theta) = \\sum_{j \\in J} I_j(\\theta)

    where :math:`J` is the set of items in the test and :math:`I_j(\\theta)` is the
    item information of :math:`j` at aspecific :math:`\\theta` value.

    :param theta: a proficiency value.
    :param items: a matrix containing item parameters.
    :returns: the test information at `theta` for a test represented by `items`.
    """
    return sum([inf(theta, item[0], item[1], item[2]) for item in items])


def reliability(theta: float, items: numpy.ndarray):
    """ Computes test reliability [Thissen00]_, given by:

    .. math:: Rel = 1 - \\frac{1}{I(\\theta)}

    Test reliability is a measure of internal consistency for the test, similar
    to Cronbach's :math:`\\alpha` in Classical Test Theory. Its value is always
    lower than 1, with values close to 1 indicating good reliability. If
    :math:`I(\\theta) < 1`, :math:`Rel < 0` and in these cases it does not make
    sense, but usually the application of additional items solves this problem.

    :param theta: a proficiency value.
    :param items: a matrix containing item parameters.
    :returns: the test reliability at `theta` for a test represented by `items`.
    """
    return 1 - (1 / test_info(theta, items))


def inf(theta: float, a: float, b: float, c: float=0) -> float:
    """Calculates the information value of an item using the Item Response Theory
    three-parameter logistic model function [Ayala2009]_:

    .. math:: I(\\theta) = a^2\\frac{(P(\\theta)-c)^2}{(1-c)^2}.\\frac{(1-P(\\theta))}{P(\\theta)}

    :param theta: the individual's proficiency value. This parameter value has
                  no boundary, but if a distribution of the form
                  :math:`N(0, 1)` was used to estimate the parameters, then
                  :math:`-4 \\leq \\theta \\leq 4`.

    :param a: the discrimination parameter of the item, usually a positive
              value in which :math:`0.8 \\leq a \\leq 2.5`.

    :param b: the item difficulty parameter. This parameter value has no
              boundary, but if a distribution of the form :math:`N(0, 1)` was
              used to estimate the parameters, then :math:`-4 \\leq b \\leq 4`.

    :param c: the item pseudo-guessing parameter. Being a probability,
        :math:`0\\leq c \\leq 1`, but items considered good usually have
        :math:`c \\leq 0.2`.

    :returns: the information value of the item at the designated `theta` point."""
    ml3 = tpm(theta, a, b, c)

    return math.pow(a, 2) * (math.pow(ml3 - c, 2) / math.pow(1 - c, 2)) * (1 - ml3) / ml3


def max_info(a=1, b=0, c=0):
    """Returns the :math:`\\theta` value to which the item with the given parameters
    gives maximum information. For the 1-parameter and 2-parameter logistic models,
    this :math:`\\theta` corresponds to where :math:`b = 0.5`. In the 3-parameter
    logistic model, however, this value is approximated by the following function:

    .. math:: argmax_{\\theta}I(\\theta) \\approx \\frac{a*b-\\log(-2(c-0.616))}{a}

    A few results can be seen in the plots below:

    .. plot::

        from catsim.cat import generate_item_bank
        from catsim import plot
        items = generate_item_bank(2)
        for item in items:
            plot.item_curve(item[0], item[1], item[2], ptype='iic', max_info=True)

    :param a: item discrimination parameter
    :param b: item difficulty parameter
    :param c: item pseudo-guessing parameter
    :param title: plot title
    :param ptype: 'icc' for the item characteristic curve, 'iic' for the item
                  information curve or 'both' for both curves in the same plot
    :param filepath: saves the plot in the given path
    :param show: whether the generated plot is to be shown

    """
    return (a * b - math.log(-2 * (c - 0.616))) / a


def logLik(est_theta: float, response_vector: list, administered_items: numpy.ndarray) -> float:
    """Calculates the log-likelihood of an estimated proficiency, given a
    response vector and the parameters of the answered items [Ayala2009]_.

    The likelihood function of a given :math:`\\theta` value given the answers to :math:`I` items is given by:

    .. math:: L(X_{Ij} | \\theta_j, a_I, b_I, c_I) = \\prod_{i=1} ^ I P_{ij}(\\theta)^{X_{ij}} Q_{ij}(\\theta)^{1-X_{ij}}

    For computational reasons, it is common to use the log-likelihood in
    maximization/minimization problems, transforming the product of
    probabilities in a sum of probabilities:

     .. math:: \\log L(X_{Ij} | \\theta_j, , a_I, b_I, c_I) = \\sum_{i=1} ^ I
               \\left\\lbrace x_{ij} \\log P_{ij}(\\theta)+ (1 - x_{ij}) \\log
               Q_{ij}(\\theta) \\right\\rbrace

    :param est_theta: estimated proficiency value.
    :param response_vector: a binary list containing the response vector.
    :param administered_items: a numpy array containing the parameters of the answered items.
    :returns: log-likelihood of a given proficiency value, given the responses to the administered items.
    """
    # inspired in the example found in
    # http://stats.stackexchange.com/questions/66199/maximum-likelihood-curve-
    # model-fitting-in-python
    # try:
    if len(response_vector) != administered_items.shape[0]:
        raise ValueError(
            'Response vector and administered items must have the same number of items'
        )
    LL = 0

    for i in range(len(response_vector)):
        prob = tpm(
            est_theta, administered_items[i][0], administered_items[i][1], administered_items[i][2]
        )

        LL += (response_vector[i] * math.log(prob)) + \
              ((1 - response_vector[i]) * math.log(1 - prob))
    return LL


def negativelogLik(est_theta: float, *args) -> float:
    """Function used by :py:mod:`scipy.optimize` optimization functions that tend to minimize
    values, instead of maximizing them. Calculates the negative log-likelihood of a proficiency
    value, given a response vector and the parameters of the administered items. The value of
    :py:func:`negativelogLik` is simply the value of :math:`-` :py:func:`logLik` or, mathematically:

    .. math:: - \\log L(X_{Ij} | \\theta_j, , a_I, b_I, c_I)

    :param est_theta: estimated proficiency value

    args:

    :param response_vector list: a binary list containing the response vector
    :param administered_items numpy.ndarray: a numpy array containing the parameters of the answered items
    :returns: negative log-likelihood of a given proficiency value, given the responses to the administered items
    """
    return -logLik(est_theta, args[0], args[1])


def normalize_item_bank(items: numpy.ndarray) -> numpy.ndarray:
    """Normalize an item matrix so that it conforms to the standard used by catsim.
    The item matrix must have dimension nx3, in which column 1 represents item discrimination,
    column 2 represents item difficulty and column 3 represents the pseudo-guessing parameter.

    If the matrix has one column, it is assumed to be the difficulty column and the other
    two columns are added such that items simulate the 1-parameter logistic model.

    If the matrix has two columns, they are assumed to be the discrimination and difficulty
    columns, respectively. the pseudo-guessing column is added such that items simulate the 2-parameter logistic model.

    :param items: the item matrix.
    :returns: an nx3 item matrix conforming to 1, 2 and 3 parameter logistic models.
    """
    if len(items.shape) == 1:
        items = numpy.expand_dims(items, axis=0)
    if items.shape[1] == 1:
        items = numpy.append(numpy.ones((items.shape[0])), items, axis=1)
    if items.shape[1] == 2:
        items = numpy.append(items, numpy.zeros((items.shape[0])), axis=1)

    return items


def validate_item_bank(items: numpy.ndarray, raise_err: bool=False):
    """Validates the shape and parameters in the item matrix so that it conforms to the standard
    used by catsim. The item matrix must have dimension nx3, in which column 1 represents item
    discrimination, column 2 represents item difficulty and column 3 represents the
    pseudo-guessing parameter.

    The item matrix must have at least one line, exactly three columns and
    :math:`\\forall i \\in I , a_i > 0 \\wedge 0 < c_i < 1`

    :param items: the item matrix.
    :param raise_err: whether to raise an error in case the validation fails or
                      just print the error message to standard output.
    """
    if type(items) is not numpy.ndarray:
        raise ValueError('Item matrix is not of type {0}'.format(numpy.ndarray))

    err = ''

    if len(items.shape) == 1:
        err += 'Item matrix has only one dimension.'
    elif items.shape[1] > 3:
        print(
            '\nItem matrix has more than 3 columns. catsim tends to add additional\
            columns to the matriz during the simulation, so it\'s not a good idea to keep them.'
        )
    elif items.shape[1] < 3:
        if items.shape[1] == 1:
            err += '\nItem matrix has no discrimination or pseudo-guessing parameter columns'
        elif items.shape[1] == 2:
            err += '\nItem matrix has no pseudo-guessing parameter column'
    else:
        if any(items[:, 2] < 0):
            err += '\nThere are items with discrimination < 0'
        if any(items[:, 2] < 0):
            err += '\nThere are items with pseudo-guessing < 0'
        if any(items[:, 2] > 1):
            err += '\nThere are items with pseudo-guessing > 1'

    if len(err) > 0 and raise_err:
        raise ValueError(err)

    print(err)



from abc import ABCMeta, abstractmethod
import numpy


class Initializer:
    """Base class for CAT initializers"""
    __metaclass__ = ABCMeta

    def __init__(self):
        super(Initializer, self).__init__()

    @abstractmethod
    def initialize(self) -> float:
        """Selects an examinee's initial :math:`\\theta` value

        :returns: examinee's initial :math:`\\theta` value
        """
        pass


class RandomInitializer(Initializer):
    """Randomly initializes the first estimate of an examinee's proficiency

    :param dist_type: either `uniform` or `normal`
    :param dist_params: a tuple containing minimum and maximum values for the
                        uniform distribution (in no particular order) or the average
                        and standard deviation values for the normal distribution
                        (in this particular order)."""

    def __init__(self, dist_type: str='uniform', dist_params: tuple=(-5, 5)):
        super(RandomInitializer, self).__init__()

        available_distributions = ['uniform', 'normal']
        if dist_type not in available_distributions:
            raise ValueError(
                '{0} not in available distributions {1}'.format(
                    dist_type, available_distributions
                )
            )

        self._dist_type = dist_type
        self._dist_params = dist_params

    def initialize(self) -> float:
        """Generates a value using the chosen distribution and parameters

        :returns: a proficiency value generated from the chosen distribution using the passed parameters"""
        if self._dist_type == 'uniform':
            return numpy.random.uniform(min(self._dist_params), max(self._dist_params))
        elif self._dist_type == 'normal':
            return numpy.random.normal(self._dist_params[0], self._dist_params[1])


class FixedPointInitializer(Initializer):
    "Initializes every proficiency at the same point."

    def __init__(self, start: float):
        super(FixedPointInitializer, self).__init__()

        self._start = start

    def initialize(self) -> float:
        """Returns the same proficiency value that was passed to the constructor of the initializer

        :returns: the same proficiency value that was passed to the constructor of the initializer"""
        return self._start



"""Module with functions for plotting IRT-related results."""

import os
import numpy
from catsim import irt
import matplotlib.pyplot as plt
from catsim.simulation import Simulator
from mpl_toolkits.mplot3d import Axes3D


def __column(matrix, i):
    """Returns columns from a bidimensional Python list (a list of lists)"""
    return [row[i] for row in matrix]


def item_curve(
    a: float=1,
    b: float=0,
    c: float=0,
    title: str=None,
    ptype: str='icc',
    max_info=True,
    filepath: str=None,
    show: bool=True
):
    """Plots 'Item Response Theory'-related item plots

    .. plot::

        from catsim.cat import generate_item_bank
        from catsim import plot
        item = generate_item_bank(1)[0]
        plot.item_curve(item[0], item[1], item[2], ptype='icc')
        plot.item_curve(item[0], item[1], item[2], ptype='iic')
        plot.item_curve(item[0], item[1], item[2], ptype='both')

    When both curves are plotted in the same figure, the figure has no grid,
    since each curve has a different scale.

    :param a: item discrimination parameter
    :param b: item difficulty parameter
    :param c: item pseudo-guessing parameter
    :param title: plot title
    :param ptype: 'icc' for the item characteristic curve, 'iic' for the item
                  information curve or 'both' for both curves in the same plot
    :param filepath: saves the plot in the given path
    :param show: whether the generated plot is to be shown
    """
    available_types = ['icc', 'iic', 'both']

    if ptype not in available_types:
        raise ValueError('\'{0}\' not in available plot types: {1}'.format(ptype, available_types))

    thetas = numpy.arange(b - 4, b + 4, .1, 'double')
    p_thetas = []
    i_thetas = []
    for theta in thetas:
        p_thetas.append(irt.tpm(theta, a, b, c))
        i_thetas.append(irt.inf(theta, a, b, c))

    if ptype in ['icc', 'iic']:
        plt.figure()

        if title is not None:
            plt.title(title, size=18)

        plt.annotate(
            '$a = ' + format(a) + '$\n$b = ' + format(
                b
            ) + '$\n$c = ' + format(c) + '$',
            bbox=dict(
                facecolor='white',
                alpha=1
            ),
            xy=(.75, .05),
            xycoords='axes fraction'
        )
        plt.xlabel(r'$\theta$')
        plt.grid()
        plt.legend(loc='best')

        if ptype == 'icc':
            plt.ylabel(r'$P(\theta)$')
            plt.plot(thetas, p_thetas)

        elif ptype == 'iic':
            plt.ylabel(r'$I(\theta)$')
            plt.plot(thetas, i_thetas)
            if max_info:
                aux = irt.max_info(a, b, c)
                plt.plot(aux, irt.inf(aux, a, b, c), 'o')

    elif ptype == 'both':
        _, ax1 = plt.subplots()

        ax1.plot(thetas, p_thetas, 'b-')
        ax1.set_xlabel(r'$\theta$', size=16)
        # Make the y-axis label and tick labels match the line color.
        ax1.set_ylabel(r'$P(\theta)$', color='b', size=16)
        for tl in ax1.get_yticklabels():
            tl.set_color('b')

        ax2 = ax1.twinx()
        ax2.plot(thetas, i_thetas, 'r-')
        ax2.set_ylabel(r'$I(\theta)$', color='r', size=16)
        for tl in ax2.get_yticklabels():
            tl.set_color('r')
        if max_info:
            aux = irt.max_info(a, b, c)
            plt.plot(aux, irt.inf(aux, a, b, c), 'o')

        if title is not None:
            ax1.set_title(title, size=18)

        ax2.annotate(
            '$a = ' + format(a) + '$\n$b = ' + format(
                b
            ) + '$\n$c = ' + format(c) + '$',
            bbox=dict(
                facecolor='white',
                alpha=1
            ),
            xy=(.75, .05),
            xycoords='axes fraction'
        )
        ax2.legend(loc='best', framealpha=0)

    if filepath is not None:
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        plt.savefig(filepath, bbox_inches='tight')

    if show:
        plt.show()


def gen3D_dataset_scatter(
    items: numpy.ndarray,
    title: str=None,
    filepath: str=None,
    show: bool=True
):
    """Generate the item matrix tridimensional dataset scatter plot

    .. plot::

        from catsim.cat import generate_item_bank
        from catsim import plot
        items = generate_item_bank(100)
        plot.gen3D_dataset_scatter(items)

    :param items: the item matrix
    :param title: the scatter plot title
    :param filepath: the path to save the scatter plot
    :param show: whether the generated plot is to be shown
    """
    irt.validate_item_bank(items)

    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(
        __column(items.tolist(), 0),
        __column(items.tolist(), 1),
        __column(items.tolist(), 2),
        s=10
    )

    if title is not None:
        plt.title(title, size=18)

    ax.set_xlabel('a')
    ax.set_ylabel('b')
    ax.set_zlabel('c')

    if filepath is not None:
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        plt.savefig(filepath, bbox_inches='tight')

    if show:
        plt.show()


def test_progress(
    title: str=None,
    simulator: Simulator=None,
    index: int=None,
    thetas: list=None,
    administered_items: numpy.ndarray=None,
    true_theta: float=None,
    info: bool=False,
    var: bool=False,
    see: bool=False,
    reliability: bool=False,
    filepath: str=None,
    show: bool=True
):
    """Generates a plot representing an examinee's test progress

    .. plot::

        from catsim.cat import generate_item_bank
        from catsim import plot
        from catsim.initialization import RandomInitializer
        from catsim.selection import MaxInfoSelector
        from catsim.estimation import HillClimbingEstimator
        from catsim.stopping import MaxItemStopper
        from catsim.simulation import Simulator

        initializer = RandomInitializer()
        selector = MaxInfoSelector()
        estimator = HillClimbingEstimator()
        stopper = MaxItemStopper(20)
        s = Simulator(generate_item_bank(100), 10)
        s.simulate(initializer, selector, estimator, stopper)
        plot.test_progress(simulator=s, index=0)
        plot.test_progress(simulator=s, index=0, true_theta=s.examinees[0], info=True, var=True, see=True)

    :param title: the plot title.
    :param simulator: a simulator which has already simulated a series of CATs,
                      containing estimations to the examinees' proficiencies and
                      a list of administered items for each examinee.
    :param index: the index of the examinee in the simulator whose plot is to be done.
    :param thetas: if a :py:class:`Simulator` is not passed, then a list of proficiency
                   estimations can be manually passed to the function.
    :param administered_items: if a :py:class:`Simulator` is not passed, then a
                               matrix of administered items, represented by their
                               parameters, can be manually passed to the function.
    :param true_theta: the value of the examinee's true proficiency. If it is passed,
                       it will be shown on the plot, otherwise not.
    :param info: plot test information. It only works if both proficiencies and
                 administered items are passed.
    :param var: plot the estimation variance during the test. It only
                works if both proficiencies and administered items are passed.
    :param see: plot the standard error of estimation during the test. It only
               works if both proficiencies and administered items are passed.
    :param reliability: plot the test reliability. It only works if both proficiencies
                        and administered items are passed.
    :param filepath: the path to save the plot
    :param show: whether the generated plot is to be shown
    """
    if simulator is None and thetas is None and administered_items is None:
        raise ValueError('Not a single plottable object was passed.')

    plt.figure()

    if title is not None:
        plt.title(title, size=18)

    if simulator is not None and index is not None:
        thetas = simulator.all_estimations[index]
        administered_items = simulator.items[simulator.administered_items[index]]
        true_theta = simulator.examinees[index]

    if thetas is not None and administered_items is not None and len(thetas) != len(
        administered_items[:, 1]
    ):
        raise ValueError(
            'Number of estimations and administered items is not the same. They should be.'
        )

    xs = range(len(thetas)) if thetas is not None else range(len(administered_items[:, 1]))

    if thetas is not None:
        plt.plot(xs, thetas, label=r'$\hat{\theta}$')
    if administered_items is not None:
        difficulties = administered_items[:, 1]
        plt.plot(xs, difficulties, label='Item difficulty')
    if true_theta is not None:
        plt.hlines(true_theta, 0, len(xs), label=r'$\theta$')
    if thetas is not None and administered_items is not None:

        # calculate and plot test information, var, standard error and reliability
        if info:
            sees = [irt.test_info(thetas[x], administered_items[:x + 1, ]) for x in xs]
            plt.plot(xs, sees, label=r'$I(\theta)$')

        if var:
            varss = [irt.var(thetas[x], administered_items[:x + 1, ]) for x in xs]
            plt.plot(xs, varss, label=r'$Var$')

        if see:
            sees = [irt.see(thetas[x], administered_items[:x + 1, ]) for x in xs]
            plt.plot(xs, sees, label=r'$SEE$')

        if reliability:
            reliabilities = [irt.reliability(thetas[x], administered_items[:x + 1, ]) for x in xs]
            plt.plot(xs, reliabilities, label='Reliability')

    plt.xlabel('Items')
    plt.grid()
    plt.legend(loc='best')

    if filepath is not None:
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        plt.savefig(filepath, bbox_inches='tight')

    if show:
        plt.show()



from abc import ABCMeta, abstractmethod
from catsim import irt
import numpy


class Stopper:
    """Base class for CAT stop criterion"""
    __metaclass__ = ABCMeta

    def __init__(self):
        super(Stopper, self).__init__()

    @abstractmethod
    def stop(self, administered_items: numpy.ndarray, estimations: list) -> bool:
        """Checks whether the test reached its stopping criterion

        :param administered_itens: a matrix with the parameters of administered Items
        :param estimations: the estimations of the examinee's proficiency so far
        :returns: `True` if the test met its stopping criterion, else `False`"""
        pass


class MaxItemStopper(Stopper):
    """Stopping criterion for maximum number of items in a test

    :param max_itens: the maximum number of items in the test"""

    def __init__(self, max_itens: int):
        super(MaxItemStopper, self).__init__()
        self._max_itens = max_itens

    def stop(self, administered_items: numpy.ndarray, estimations: list) -> bool:
        """Checks whether the test reached its stopping criterion

        :param administered_itens: a matrix with the parameters of administered Items
        :param estimations: the estimations of the examinee's proficiency so far
        :returns: `True` if the test met its stopping criterion, else `False`"""
        n_itens = administered_items.shape[0]
        if n_itens > self._max_itens:
            raise ValueError(
                'More items than permitted were administered: {0} > {1}'.format(
                    n_itens, self._max_itens
                )
            )
        return n_itens == self._max_itens


class MinErrorStopper(Stopper):
    """Stopping criterion for minimum standard error of estimation (see :py:func:`catsim.irt.see`)

    :param min_error: the minimum standard error of estimation the test must achieve before stopping"""

    def __init__(self, min_error: float):
        super(MinErrorStopper, self).__init__()
        self._min_error = min_error

    def stop(self, administered_items: numpy.ndarray, estimations: list) -> bool:
        """Checks whether the test reached its stopping criterion

        :param administered_itens: a matrix with the parameters of administered Items
        :param estimations: the estimations of the examinee's proficiency so far
        :returns: `True` if the test met its stopping criterion, else `False`"""
        if len(estimations) == 0:
            return False

        theta = estimations[-1]
        return irt.see(theta, administered_items) < self._min_error



import unittest
import numpy
from catsim import irt
from catsim import cat
from catsim import plot
from catsim import stats
from catsim.simulation import Simulator
from catsim.cat import generate_item_bank
from catsim.stopping import MaxItemStopper, MinErrorStopper
from catsim.selection import MaxInfoSelector, ClusterSelector, LinearSelector, RandomSelector, AStratifiedBBlockingSelector, MaxInfoStratificationSelector, MaxInfoBBlockingSelector, AStratifiedSelector, The54321Selector, RandomesqueSelector
from catsim.initialization import RandomInitializer, FixedPointInitializer
from catsim.estimation import HillClimbingEstimator, DifferentialEvolutionEstimator, FMinEstimator


def test_item_bank_generation():
    for items in [
        generate_item_bank(5, '1PL'), generate_item_bank(5, '2PL'), generate_item_bank(5, '3PL'),
        generate_item_bank(
            5,
            '3PL',
            corr=0
        )
    ]:
        irt.validate_item_bank(items, raise_err=True)

    items = numpy.zeros((100))
    irt.validate_item_bank(items)
    items = irt.normalize_item_bank(items)
    irt.validate_item_bank(items, raise_err=True)


def test_plots():
    from matplotlib.pyplot import close
    initializer = RandomInitializer()
    selector = MaxInfoSelector()
    estimator = HillClimbingEstimator()
    stopper = MaxItemStopper(20)
    s = Simulator(generate_item_bank(100), 10)
    s.simulate(initializer, selector, estimator, stopper)

    for item in s.items[0:10]:
        yield plot.item_curve, item[0], item[1], item[2], 'Test plot', 'icc', False, None, False
        yield plot.item_curve, item[0], item[1], item[2], 'Test plot', 'iic', True, None, False
        yield plot.item_curve, item[0], item[1], item[2], 'Test plot', 'both', True, None, False
        close('all')

    plot.gen3D_dataset_scatter(items=s.items, show=False)
    plot.test_progress(
        title='Test progress',
        simulator=s,
        index=0,
        info=True,
        see=True,
        reliability=True,
        show=False
    )

    # close all plots after testing
    close('all')


def test_stats():
    import numpy.random as nprnd
    for _ in range(10):
        items = generate_item_bank(500)
        stats.coef_variation(items)
        stats.coef_correlation(items)
        stats.covariance(items)
        stats.covariance(items, False)
        stats.scatter_matrix(items)

        random_integers = nprnd.randint(30, size=1000)
        stats.bincount(random_integers)


def one_simulation(items, examinees, initializer, selector, estimator, stopper):
    s = Simulator(items, examinees)
    s.simulate(initializer, selector, estimator, stopper)
    cat.rmse(s.examinees, s.estimations)


def test_simulations():
    examinees = 10
    test_size = 20
    bank_size = 5000

    logistic_models = [
        # '1PL', '2PL',
        '3PL'
    ]
    initializers = [
        RandomInitializer('uniform',
                          (-5, 5)
                          # ), RandomInitializer(
                          #     'normal', (0, 1)
                          ),
        FixedPointInitializer(0)
    ]
    infinite_selectors = [MaxInfoSelector(), RandomSelector()]
    finite_selectors = [
        LinearSelector(
            numpy.random.randint(
                bank_size,
                size=test_size
            )
        ), AStratifiedSelector(test_size), AStratifiedBBlockingSelector(test_size),
        MaxInfoStratificationSelector(test_size), MaxInfoBBlockingSelector(test_size),
        The54321Selector(test_size), RandomesqueSelector(5)
    ]

    estimators = [HillClimbingEstimator(), DifferentialEvolutionEstimator((-8, 8)), FMinEstimator()]

    for logistic_model in logistic_models:
        for initializer in initializers:
            for estimator in estimators:
                for stopper in [MaxItemStopper(test_size)]:
                    for selector in finite_selectors:
                        items = generate_item_bank(bank_size, itemtype=logistic_model)
                        yield one_simulation, items, examinees, initializer, selector, estimator, stopper

                for stopper in [MinErrorStopper(.4)]:
                    for selector in infinite_selectors:
                        items = generate_item_bank(bank_size, itemtype=logistic_model)
                        yield one_simulation, items, examinees, initializer, selector, estimator, stopper


def test_cism():
    from sklearn.cluster import KMeans

    examinees = 10
    initializers = [
        RandomInitializer('uniform',
                          (-5, 5)
                          # ), RandomInitializer(
                          #     'normal', (0, 1)
                          ),
        FixedPointInitializer(0)
    ]
    estimators = [HillClimbingEstimator(), DifferentialEvolutionEstimator((-8, 8)), FMinEstimator()]
    stoppers = [MaxItemStopper(20), MinErrorStopper(.4)]

    for initializer in initializers:
        for estimator in estimators:
            for stopper in stoppers:
                items = generate_item_bank(5000)
                clusters = list(KMeans(n_clusters=8).fit_predict(items))
                ClusterSelector.weighted_cluster_infos(0, items, clusters)
                ClusterSelector.avg_cluster_params(items, clusters)
                selector = ClusterSelector(clusters=clusters, r_max=.2)
                yield one_simulation, items, examinees, initializer, selector, estimator, stopper


if __name__ == '__main__':
    unittest.main()
